unit dnPxGrid;

(*****************************************************************************
Copyright  1997-99 Dieter Nth
version    : 2.1
date       : 10.08.1999
description:

  Component TPxGrid:
    This is an enhanced TStringGrid, which adds support for Paradox-style masks

    new property:
      PxMask (public)
    new event:
      OnValidChanged (published);

    It still uses Borland's EditMask (so PLEASE DO NOT SUPPLY BOTH EditMask and
    PxMask) and validates it.

******************************************************************************)

interface

{$I dnInc.INC}

uses
  Windows, SysUtils, Grids, Controls, Graphics, Classes, Messages, Forms,
  Consts,
  Mask, Dialogs,
  dnPxMask;

type
  TGridPxMask = class(TControlPxMask)
  private
    procedure CMExit(var Message: TCMExit); message CM_EXIT;
    procedure WMSetFocus(var Message: TWMSetFocus); message WM_SETFOCUS;
  protected
    function CanEditModify: boolean; override;
    function GetEditHandle: THandle; override;
    procedure Reset; override;
  public
    procedure HookWndProc(var Message: TMessage); override;
    procedure ValidateEdit; override;
    procedure PrepareShowError;
    function SelectCell(ACol, ARow: Longint): boolean;
  published
    property OnValidChanged stored false;
  end;

  TPxGrid = class(TStringGrid)
  private
    FPxMask: TGridPxMask;
    function GetOnValidChanged: TValidChangedEvent;
    procedure SetOnValidChanged(value: TValidChangedEvent);
    procedure SetPxMask(value: TGridPxMask);
  protected
    procedure KeyPress(var Key: Char); override;
    function SelectCell(ACol, ARow: Longint): boolean; override;
  public
    constructor Create(aOwner: TComponent); override;
    destructor Destroy; override;
  published
    property PxMask: TGridPxMask read FPxMask write SetPxMask stored false;
      //must be published
    property OnValidChanged: TValidChangedEvent read GetOnValidChanged
      write SetOnValidChanged;
  end;

implementation

{*** TGridPxMask ***}
procedure TGridPxMask.CMExit(var Message: TCMExit);
begin
  try
    ValidateEdit;
    inherited;
    exclude(FState, pxReEnter);
  except
    PrepareShowError;
    raise;
  end;
end;

procedure TGridPxMask.WMSetFocus(var Message: TWMSetFocus);
var
  wasEditModified: boolean;
begin
  wasEditModified := (Message.FocusedWnd = 0) and
    assigned(TPxGrid(Owner).InplaceEditor) and
    EditModified and not TPxGrid(Owner).InplaceEditor.visible;
  if wasEditModified then
    include(FState, pxReEnter);
  inherited;
  if wasEditModified then
    exclude(FState, pxReEnter);
end;

function TGridPxMask.CanEditModify: boolean;
begin
  result := TPxGrid(Owner).CanEditModify;
  with TPxGrid(Owner) do
  begin
    if result and not EditorMode then
    begin
      EditorMode := true;
      if assigned(InplaceEditor) then InplaceEditor.SelectAll;
    end;
  end;
end;

function TGridPxMask.GetEditHandle: THandle;
begin
  result := 0;
  if assigned(TPxGrid(Owner).InplaceEditor) then
    result := TPxGrid(Owner).InplaceEditor.Handle;
end;

procedure TGridPxMask.HookWndProc(var Message: TMessage);
begin
  with Message do
  begin
    case Msg of
      WM_KillFocus:
        with TWMSetFocus(Message) do
        begin
          if (FocusedWnd <> 0) and (FocusedWnd <> TPxGrid(Owner).Handle) and
            assigned(TPxGrid(Owner).InplaceEditor) and
            (FocusedWnd <> TPxGrid(Owner).InplaceEditor.Handle) and
            EditModified and not TPxGrid(Owner).InplaceEditor.visible then
          begin
            try
              if AutoValidate then ValidateEdit;
              exclude(FState, pxReEnter);
            except
              PrepareShowError;
              raise;
            end;
          end
        end;
    end;
  end;
  inherited;
end;

procedure TGridPxMask.Reset;
begin
  inherited;
  with TPxGrid(Owner).InplaceEditor do
    if isMasked then
      Text := EditText;
end;

procedure TGridPxMask.ValidateEdit;
var
  OldMask: Char;
begin
  with TPxGrid(Owner) do
    if assigned(InplaceEditor) then
    begin
      try
        InplaceEditor.ValidateEdit;
      except
        PrepareShowError;
        raise;
      end;

      inherited;

      exclude(FState, pxReEnter);
      OldMask := MaskNoSave;
      MaskNoSave := #0;
      SetEditText(Col, Row, InplaceEditor.Text);
      MaskNoSave := OldMask;
    end;
end;

procedure TGridPxMask.PrepareShowError;
var
  OldText: string;
  OldSelStart: integer;
begin
  with TPxGrid(Owner) do
  begin
    OldSelStart := InplaceEditor.SelStart;
    OldText := InplaceEditor.EditText;
    if (Col < LeftCol) or (Col > LeftCol + VisibleColCount) or
      (Row < TopRow) or (Row > TopRow + VisibleRowCount) then
    try
      include(FState, pxReEnter);
      include(FState, pxNoValidation);
      MoveColRow(Col, Row, false, true);
    finally
      exclude(FState, pxNoValidation);
    end;
    if CanFocus then SetFocus;
    ShowEditor;
    InplaceEditor.EditText := OldText;
    if InplaceEditor.isMasked then
    begin
      InplaceEditor.SelStart := OldSelStart;
      InplaceEditor.SelLength := 1;
    end
    else
      SetCaret;
    InplaceEditor.Modified := true;
    EditModified := true;
  end;
end;

function TGridPxMask.SelectCell(ACol, ARow: Longint): boolean;
begin
  result := true;
  with TPxGrid(Owner) do
  begin
    if EditorMode then
    begin
      if not (pxNoValidation in FState) then
      try
        if AutoValidate then ValidateEdit;
        EditModified := false;
        exclude(FState, pxReEnter);
      except
        PrepareShowError;
        Application.HandleException(self);
        result := false;
        exit;
      end;
    end;
    PrepareReset(GetEditText(ACol, ARow));
  end;
end;

{*** TPxGrid ***}

constructor TPxGrid.Create(aOwner: TComponent);
begin
  inherited Create(aOwner);
  FPxMask := TGridPxMask.Create(self, true); //MUST be created after inherited
end;

destructor TPxGrid.Destroy;
begin
  if assigned(FPxMask) then FPxMask.free;
  inherited Destroy;
end;

function TPxGrid.GetOnValidChanged: TValidChangedEvent;
begin
  result := FPxMask.OnValidChanged;
end;

procedure TPxGrid.SetOnValidChanged(value: TValidChangedEvent);
begin
  FPxMask.OnValidChanged := value;
end;

procedure TPxGrid.SetPxMask(value: TGridPxMask);
begin
  FPxMask.assign(value);
end;

procedure TPxGrid.KeyPress(var Key: Char);
begin
  FPxMask.CheckResetKey(Key);
  FPxMask.HookKeyPress(Key);
  inherited KeyPress(Key);
end;

function TPxGrid.SelectCell(ACol, ARow: Longint): boolean;
begin
  result := inherited SelectCell(ACol, ARow) and FPxMask.SelectCell(ACol, ARow);
end;

end.
