unit dnPxMask;

(*****************************************************************************
Copyright  1997-99 Dieter Nth
version    : 2.1
date       : 10.08.1999
description:

  This unit implements Paradox-style Inputmasks:
  most of the validation-code of Inputmasks is based upon Borland's VALIDATE.PAS
  Unit ( 1992 Borland International) from Borland Pascal 7,
  i don't know why they dropped it from VCL...

Bugs fixed:
1.03 :
  Borland's masks have a 'feature', which caused errors:
  there may be empty choices in groups (without meaning), e.g. ,,
  or *[&[*?]*[{#, ,.,;,,;;,:,;[,;],(,),}]]
                                      ^ <-- Paradox works fine with this, but my
  PxMask crashed sometimes. I couldn't fix this, so i decided to remove all
  unnecessary commas, when setting the mask :-)
1.02 :
  Borland's example input mask (from DBD)
  *[&[*?]*[{#, ,.,;,,;;,:,;[,;],(,),}]]  (capitalized words) allowed only
  ONE word to be entered.
  Yes, I should have tried at least the example masks, but I did it once and
  then I found a bug in Borland's Iteration function causing infinite loops in
  iterations (without a number) containing groups....
  I thought, I fixed it, but this fix caused this new problem...
  There still may be some combinations of groups and iterations that will
  cause an infinite loop. So, if your running your program from within Delphi
  and it hangs while typing in a TPxMaskEdit, set a breakpoint in
  function Iteration and stop your program. If this happens within the
  PxMask-property editor you'll have to kill Delphi ;-).
  Before You send me an email complaining about this, please try to follow
  all those recursive calls with debugger (and remember, that part of code
  wasn't written by Borland, not me), maybe YOU'll fix it. If You don't,
  I'll try my best.

Changes:
2.00:
  Put most of the PxMask-code in a class, so it's easier to maintain the code and
  create components with a PxMask, BUT now the properties changed:
    PxMask is now an CLASS-property NOT a string: AutoFill, AutoValidate,
    RaiseInputError, PxMask, ErrorMessage are now properties of that class.
  So you'll have to change your project-source, sorry for the inconvenience...

1.03:
  Changed strings in english RES-file to be equivalent to those in DBD32:
    SPxMaskIllegalString, "The field value fails picture validity check"
    SPxMaskIllegalChar, "Illegal character for picture validation"

  added new error string, if input string is valid, but not yet complete:
    SPxMaskInputIncomplete, "Input OK, but not complete"

  new property ErrorText: string;
    You may include the PxMask by inserting '%MASK%', which is replaced with
    the actual PxMask, e.g. 'Input does not confirm to picture: %MASK%'.
    This means, if you want to make this the standard behaviour, you may change
    the default-strings SPxMaskIllegalString/SPxMaskIllegalChar in the RES-file.
    You may also include %ERROR% in a customised ErrorMessage, this includes the
    original ErrorMessage.

1.02 :
  Testing an input mask in DBD32 I noticed, that there's a new special char:
  ---> ~ <--- lower case Char
  and, of course, I implemented it.
  If You don't want to use this new feature, you may comment out that part in
  function Scan

******************************************************************************)

{$I dnInc.INC}

interface

uses
  Windows, Messages, SysUtils, Classes, Controls, StdCtrls, Forms, Menus,
  Consts, Mask,
  dnPxCnst, dnGlobal;

const
  errPxMaskSyntax = 1;
  errPxMaskIllegalChar = 2;
  errPxMaskInputIncomplete = 3;
  errPxMaskIllegalString = 4;

type
EPxMaskError = class(EDBEditError)
  public
    ErrorCode : integer;
  end;

  TPxMaskState = set of ( pxReEnter, pxSending, pxValid, pxModified,
                          pxNoValidation, pxStored, pxMaskFromParadox );

{*** Base class for handling Paradox masks***}
  TPxMask = class(TPersistent)
  private
    FMask: string;
    FErrMsg: string;
    FErrPos : integer;
    FAutoFill,
    FAutoValidate,
    FRaiseInputError : boolean;

    procedure SetAutoValidate(value: boolean);
    procedure SetAutoFill(value: boolean);
    procedure SetErrMsg(const value: string);
    procedure SetErrorPos(value : integer);
    procedure SetMask(const value : string);
    procedure SetRaiseInputError(value: boolean);

    function MaskStored : boolean;
    function PropsStored: boolean;

    function CheckValid(s : string): string;
    function FormatErrorText(Ident: string): string;
  protected
    function MaskLoaded : boolean; virtual;

    procedure ValidChanged; virtual;

    {*** raises EDBEditError, if Prop RaiseInputError = true,
         gets ErrorText via FormatErrorText(vePxMaskIllegalChar)}
    procedure InputError; dynamic;
    {*** raises EPxMaskError}
    procedure ValidateError(Incomplete : boolean); dynamic;

  public
    FState : TPxMaskState;
    constructor create(aOwner : TComponent);

    procedure Assign(Source: TPersistent); override;

    {*** Checks value with func PxMaskValidate}
    procedure Validate(var value : string);

    procedure LoadMask;

    procedure Clear;
    function Cleared: boolean;

    {*** raises EPxMaskError, if aMask contains syntaxerrors}
    procedure CheckSyntax(var aMask : string);

    {*** true if PxMask <> ''}
    function isMasked : boolean;

    {*** Position of last Error after any validation}
    property ErrorPos : integer read FErrPos write SetErrorPos;

  published
    {*** Custom error message}
    property ErrorMessage : string     read FErrMsg
                                       write SetErrMsg
                                       stored PropsStored;

    property Mask : string             read FMask
                                       write SetMask
                                       stored MaskStored;

    property AutoValidate : boolean    read FAutoValidate
                                       write SetAutoValidate
                                       stored PropsStored
                                       default AutoValidateDefault;

    property AutoFill : boolean        read FAutoFill
                                       write SetAutoFill
                                       stored PropsStored
                                       default AutoFillDefault;

    property RaiseInputError : boolean read FRaiseInputError
                                       write SetRaiseInputError
                                       stored PropsStored
                                       default RaiseInputErrorDefault;
  end;

  TControlPxMask = class;

  TValidChangedEvent = procedure( Sender: TControlPxMask;
                                  EditControl: TWinControl;
                                  TextValid: boolean) of object;

  TControlPxMask = class(TPxMask)
  private
    FOwner: TComponent;
    FControlWndProc: TWndMethod;
    FOnValidChanged: TValidChangedEvent;
    FOldValue: String;
    FEnableReset: boolean;

    function GetText: string;
    procedure SetText(const value: string);
    function GetMaxLen: integer;
    function GetTextLen: integer;
    function GetEditControl: TWinControl;
    function GetModified : boolean;
    procedure CheckEditText;
  protected
    procedure PrepareReset(const OldText : string); dynamic;
    procedure Reset; dynamic;
    function CanEditModify: boolean; dynamic;

    {*** Used internally: TCustomEdit --> result = Owner.Handle
                          TCustomComboBox --> result = Owner.EditHandle
                          TCustomGrid --> result = Owner.InplaceEditor.Handle }
    function GetEditHandle: THandle; virtual;
    procedure SetModified(value: boolean); virtual;
    procedure ValidChanged; override;
    procedure ValidateError(Incomplete : boolean); override;
  public
    constructor create(aOwner : TComponent; EnableReset: boolean);
    destructor Destroy; override;

    {*** To be called from Owner's KeyPress}
    procedure HookKeyPress(var Key: Char);
    procedure CheckResetKey(var Key : char);
    procedure SetCaret;

    procedure ValidateEdit; dynamic;

    {*** Replaces Owner's WindowProc}
    procedure HookWndProc(var Message: TMessage); virtual;

    procedure DefaultHandler(var Message); override;
    property EditModified: boolean read GetModified write SetModified;
    property EditText : string read GetText write SetText;
    property EditTextLen: integer read GetTextLen;
    property EditMaxLen: integer read GetMaxLen;
    property EditHandle: THandle read GetEditHandle;
    property EditControl: TWinControl read GetEditControl;
    property Owner : TComponent read FOwner;
  published
    property OnValidChanged: TValidChangedEvent read FOnValidChanged
                                                write FOnValidChanged;
  end;

  TPxMaskEdit = class(TEdit)
  private
    FPxMask : TControlPxMask;
    function GetOnValidChanged: TValidChangedEvent;
    procedure SetOnValidChanged(value: TValidChangedEvent);
    procedure SetPxMask(value : TControlPxMask);
 protected
  public
    constructor create(aOwner : TComponent); override;
    destructor Destroy; override;
  published
    property PxMask : TControlPxMask read FPxMask write SetPxMask;
    property OnValidChanged: TValidChangedEvent read GetOnValidChanged
                                                write SetOnValidChanged
                                                stored false;
  end;

  TPxComboBox = class(TComboBox)
  private
    FPxMask : TControlPxMask;
    function GetOnValidChanged: TValidChangedEvent;
    procedure SetOnValidChanged(value: TValidChangedEvent);
    procedure SetPxMask(value : TControlPxMask);
  protected
    procedure KeyPress(var Key: Char); override;
  public
    constructor create(aOwner : TComponent); override;
    destructor Destroy; override;
  published
    property PxMask : TControlPxMask read FPxMask write SetPxMask;
    property OnValidChanged: TValidChangedEvent read GetOnValidChanged
                                                write SetOnValidChanged
                                                stored false;
  end;

{*** check syntax of supplied mask, if an exception is raised then ErrorPos > 0;
     removes any trailing blanks and illegal commas: '[a,b,,c]  ' -> '[a,b,c]' }
procedure PxMaskCheckSyntax(var aMask: string; var ErrorPos: integer);

{*** aMask : ParadoxMask (no syntax checks)
     Input : string to be checked
     AutoFill : if true then Input will automatically be completed
     returns position of error in Input, if any: 0 = OK}
function PxMaskCheckInput(const aMask: string; var Input: string;
  AutoFill: Boolean): integer;

{*** aMask : ParadoxMask (no syntax checks)
     Input : string to be checked
     returns position of error in Input, or length(Input), if not complete}
function PxMaskValidate(const aMask: string; const Input: string): integer;

{*** returns TRUE if Control has a published property PxMask}
function HasPxMask(Control: TPersistent; var PxMask: TPxMask) : boolean;

implementation
uses TypInfo,
     dnUtil;

{*** copied from VALIDATE.PAS (and partially modified)}
type
  TPxMaskResult = (prComplete, prIncomplete, prEmpty, prError, prSyntax,
    prAmbiguous, prIncompNoFill);

const
  SpecialChars = ['#', '?', '&', '!', '@', '*', '{', '}', '[', ']', ',', '~'];
  Digits = ['0' .. '9'];

function IsComplete(Rslt: TPxMaskResult): Boolean;
begin
  result := Rslt in [prComplete, prAmbiguous];
end;

function IsIncomplete(Rslt: TPxMaskResult): Boolean;
begin
  result := Rslt in [prIncomplete, prIncompNoFill];
end;

function UpCase(c : char) : char;
begin
  result := char(longInt(ansiUpper(pChar(longInt(c)))));
end;

function LoCase(c : char) : char;
begin
  result := char(longInt(ansiLower(pChar(longInt(c)))));
end;

{The compiler produces following warnings/hints:
  Hint - Value assigned to 'Iteration' never used
  Warning - Variable 'Rslt' might not have been initialized.
  Warning - Variable 'IncompJ' might not have been initialized.
  Warning - Variable 'Incomp' might not have been initialized.

  They're produced from Borland's original code, there wasn't such a smart
  compiler in BP7, that's why I turn off warnings and hints for function Picture
}

{$WARNINGS Off}
{$HINTS Off}

function Picture(const aMask: string; var Input: string;
  AutoFill: Boolean; var ErrorPos : integer): TPxMaskResult;
var
  I, J: cardinal;
  Rslt: TPxMaskResult;
  Reprocess: Boolean;

  function Process(TermCh: cardinal): TPxMaskResult;
  var
    Rslt: TPxMaskResult;
    Incomp: Boolean;
    OldI, OldJ, IncompJ, IncompI: cardinal;

    { Consume input }

    procedure Consume(Ch: Char);
    begin
      Input[J] := Ch;
      Inc(J);
      Inc(I);
      inc(ErrorPos);
    end;

    { Skip a character or a picture group }

    procedure ToGroupEnd(var I: cardinal);
    var
      BrkLevel, BrcLevel: Integer;
    begin
      BrkLevel := 0;
      BrcLevel := 0;
      repeat
        if I = TermCh then Exit;
        case aMask[I] of
          '[': Inc(BrkLevel);
          ']': Dec(BrkLevel);
          '{': Inc(BrcLevel);
          '}': Dec(BrcLevel);
          ';': Inc(I);
          '*':
            begin
              Inc(I);
              while aMask[I] in Digits do Inc(I);
              ToGroupEnd(I);
              Continue;
            end;
        end;
        Inc(I);
      until (BrkLevel = 0) and (BrcLevel = 0);
    end;

    { Find the a comma separator }

    function SkipToComma: Boolean;
    begin
      repeat ToGroupEnd(I) until (I = TermCh) or (aMask[I] = ',');
      while (I <= TermCh) and (aMask[I] = ',') do Inc(I);
      result := I < TermCh;
    end;

    { Calclate the end of a group }

    function CalcTerm: cardinal;
    begin
      result := I;
      ToGroupEnd(result);
    end;

    { The next group is repeated X times }

    function Iteration: TPxMaskResult;
    var
      oldJ, Itr, K, L: cardinal;
      NewTermCh: cardinal;
      Rslt: TPxMaskResult;
    begin
      Itr := 0;
      Iteration := prError;

      Inc(I);  { Skip '*' }

      { Retrieve number }

      while aMask[I] in Digits do
      begin
        Itr := Itr * 10 + Byte(aMask[I]) - Byte('0');
        Inc(I);
      end;

      if I > TermCh then
      begin
        Iteration := prSyntax;
        Exit;
      end;

      K := I;
      NewTermCh := CalcTerm;

      { If Itr is 0 allow any number, otherwise enforce the number }
      if Itr <> 0 then
      begin
        for L := 1 to Itr do
        begin
          I := K;
          Rslt := Process(NewTermCh);
          if not IsComplete(Rslt) then
          begin
            { Empty means incomplete since all are required }
            (**** BugFix: empty means incomplete ONLY if no optional input.
                 e.g. *3[..] means 0 to 3 times, *3{...} means exactly 3 times *)
            if (Rslt = prEmpty) then
            begin
              if (aMask[K] <> '[') then
                Iteration := prIncomplete
              else
                Iteration := prAmbiguous;
            end;
            exit;
          end;
        end;
      end
      else
      begin
        repeat
          OldJ := J;
          I := K;
          Rslt := Process(NewTermCh);
        until not IsComplete(Rslt) or (OldJ = J);
        if (Rslt = prEmpty) or (Rslt = prError) then
        begin
          Inc(I);
          Rslt := prAmbiguous;
        end;
      end;
      if IsComplete(Rslt) then I := NewTermCh;
      Iteration := Rslt;
    end;

    { Process a picture group }

    function Group: TPxMaskResult;
    var
      TermCh: cardinal;
    begin
      TermCh := CalcTerm;
      Inc(I);
      Result := Process(TermCh - 1);
      if not IsIncomplete(Result) then I := TermCh;
    end;

    function CheckComplete(Rslt: TPxMaskResult): TPxMaskResult;
    var
      J, Z: cardinal;
    begin
      J := I;
      if IsIncomplete(Rslt) then
      begin
        { Skip optional pieces }
        while True do
          case aMask[J] of
            '[': ToGroupEnd(J);
            '*':
              if not (aMask[J + 1] in Digits) then
              begin
                Inc(J);
                ToGroupEnd(J);
              end
              else
              begin
                Z := J;
                repeat inc(Z) until not (aMask[Z] in Digits);
                if aMask[Z] = '[' then
                begin
                  J := Z;
                  ToGroupEnd(J);
                end
                else
                  Break;
              end;
          else
            Break;
          end;

        if J = TermCh then Rslt := prAmbiguous;
      end;
      Result := Rslt;
    end;

    function Scan: TPxMaskResult;
    var
      Ch: Char;
      Rslt: TPxMaskResult;
    begin
      Scan := prError;
      Rslt := prEmpty;
      while (I <> TermCh) and (aMask[I] <> ',') do
      begin
        if J > Length(Input) then
        begin
          Scan := CheckComplete(Rslt);
          Exit;
        end;

        Ch := Input[J];
        case aMask[I] of
          '#': if not (ch in Digits) then Exit
               else Consume(Ch);
          '?': if not isCharAlpha(ch) then Exit
               else Consume(Ch);
          '&': if not isCharAlpha(ch) then Exit
               else Consume(UpCase(ch));
          {New special char in DBD32}
          '~': if not isCharAlpha(ch) then Exit
               else Consume(LoCase(ch));
          '!': Consume(UpCase(ch));
          '@': Consume(Ch);
          '*':
            begin
              Rslt := Iteration;
              if not IsComplete(Rslt) then
              begin
                Scan := Rslt;
                Exit;
              end;
              if Rslt = prError then Rslt := prAmbiguous;
            end;
          '{':
            begin
              Rslt := Group;
              if not IsComplete(Rslt) then
              begin
                Scan := Rslt;
                Exit;
              end;
            end;
          '[':
            begin
              Rslt := Group;
              if IsIncomplete(Rslt) then
              begin
                Scan := Rslt;
                Exit;
              end;
              if (Rslt = prError) then Rslt := prAmbiguous;
            end;
        else
          if aMask[I] = ';' then Inc(I);
         if (UpCase(aMask[i]) = UpCase(ch)) or (Ch = ' ')
          then Consume(aMask[I])
          else Exit;
        end;

        if Rslt = prAmbiguous then
          Rslt := prIncompNoFill
        else
          Rslt := prIncomplete;
      end;

      if Rslt = prIncompNoFill then
        Scan := prAmbiguous
      else
        Scan := prComplete;
    end;

  begin
    Incomp := False;
    OldI := I;
    OldJ := J;
    repeat
      Rslt := Scan;

      { Only accept completes if they make it farther in the input
        stream from the last incomplete }
      {Bug???Fix??? J < IncompJ --> J <= IncompJ}
      if (Rslt in [prComplete, prAmbiguous]) and Incomp and (J <= IncompJ) then
      begin
        Rslt := prIncomplete;
        J := IncompJ;
      end;

      if (Rslt = prError) or (Rslt = prIncomplete) then
      begin
        Process := Rslt;
        if not Incomp and (Rslt = prIncomplete) then
        begin
          Incomp := True;
          IncompI := I;
          IncompJ := J;
        end;
        I := OldI;
        J := OldJ;
        if not SkipToComma then
        begin
          if Incomp then
          begin
            Process := prIncomplete;
            I := IncompI;
            J := IncompJ;
          end;
          Exit;
        end;
        OldI := I;
      end;
    until (Rslt <> prError) and (Rslt <> prIncomplete);

    if (Rslt = prComplete) and Incomp then
      Process := prAmbiguous
    else
      Process := Rslt;
  end;
begin
  ErrorPos := 1;
  Picture := prComplete;
  if aMask = '' then exit;

  Picture := prEmpty;
  if Input = '' then Exit;

  J := 1;
  I := 1;

  Rslt := Process(Length(aMask) + 1);
  if (Rslt <> prError) and (Rslt <> prSyntax) and (J <= Length(Input)) then
    Rslt := prError;

  if (Rslt = prIncomplete) and AutoFill then
  begin
    Reprocess := False;
    while (I <= Length(aMask)) and
      not (aMask[I] in SpecialChars) do
    begin
      if aMask[I] = ';' then Inc(I);
      Input := Input + aMask[I];
      Inc(I);
      Reprocess := True;
    end;
    J := 1;
    I := 1;
    if Reprocess then
      Rslt := Process(Length(aMask) + 1)
  end;

  case Rslt of
    prAmbiguous : Picture := prComplete;
    prIncompNoFill : Picture := prIncomplete;
  else
    Picture := Rslt;
  end;
end;

{$WARNINGS On}
{$HINTS On}

function RemoveCommas(const aMask : string): string;
type
  TStatus = (none, wasLiteral, noComma);
var
  len, i, j : integer;
  status : TStatus;

  procedure PutChar;
  begin
    inc(j);
    if i <= length(aMask) then result[j] := aMask[i];
    inc(i);
  end;

begin
  len := Length(aMask);
  result := aMask;
  setLength(result, len);
  j := 0;
  i := 1;
  status := noComma;
  while i <= len do
  begin
    case aMask[i] of
      '*':
        begin
          PutChar;
          while aMask[i] in Digits do PutChar;
          status := noComma;
          continue;
        end;
      '[', '{' : status := noComma;
      ']', '}' :
        begin
          if (status <> wasLiteral) and (result[j] = ',') then dec(j);
          status := none;
        end;
      ';':
        begin
          PutChar;
          status := wasLiteral;
        end;
      ',':
        begin
          if status = noComma then
          begin
            while ((i <= len) and (aMask[i] = ',')) do inc(i);
            status := noComma;
            continue;
          end;
          status := noComma;
        end;
      else
        status := none;
    end;
    PutChar;
  end;
  if (status <> wasLiteral) and (aMask[len] = ',') then dec(j);
  setLength(result, j);
end;

procedure PxMaskCheckSyntax(var aMask : string; var ErrorPos : integer);
const
  MaxBracketLevel = 32;  { maximale Anzahl der Klammerebenen }
var
  i: Integer;
  fStack : array[0..pred(MaxBracketLevel)] of char; {Klammerstack}
  fBracketLevel : shortInt;
  isLiteral : boolean;

  procedure RaiseError(const Err : string);
  var
    Error : EPxMaskError;
  begin
    ErrorPos := i;
    Error := EPxMaskError.create(SPxMaskSyntax+ #13#10 + Err);
    Error.ErrorCode := errPxMaskSyntax;
    raise Error;
  end;

  procedure PopBracket;
  begin
    if fBracketLevel > 0 then
    begin
      dec(fBracketLevel);
      if (fStack[fBracketLevel] <> aMask[i]) then
        RaiseError(Format(SPxMaskBracketError, [fStack[fBracketLevel]]));
    end else
      raiseError(Format(SPxMaskBracketClose, [aMask[i]]));
  end;

  procedure PushBracket(c : char);
  begin
    if fBracketLevel >= MaxBracketLevel then
      raiseError(SPxMaskBracketLevel);
    fStack[fBracketLevel] := c;
    inc(fBracketLevel);
  end;
begin
  ErrorPos := 0;
  aMask := TrimRight(aMask);
  if aMask = '' then Exit;
  fBracketLevel := 0;
  i := 1;
  isLiteral := false;
  while i <= Length(aMask) do
  begin
    case aMask[i] of
      '[': pushBracket(']');
      '{': pushBracket('}');
      ']', '}':
        begin
          popBracket;
          case aMask[i-1] of
            '[', '{':
              if not isLiteral then
                raiseError(SPxMaskBracketEmpty);
          end;
        end;
      ';':
        begin
          Inc(i, 2);
          isLiteral := true;
          continue;
        end;
      '*':
        begin
          inc(i);
          while aMask[i] in Digits do inc(i);
          if (i > length(aMask)) or (aMask[i] in [']', '}', '*']) then
            RaiseError(SPxMaskIterationExpected);
          continue;
        end;
    end;
    isLiteral := false;
    Inc(i);
  end;
  if fBracketLevel > 0 then
  begin
    dec(fBracketLevel);
    RaiseError(Format(SPxMaskBracketMissing, [fStack[fBracketLevel]]));
  end;
  i := length(aMask);
  if (aMask[i] = ';') and not isLiteral then
    RaiseError(SPxMaskLiteralExpected);
  aMask := RemoveCommas(aMask);
end;

function PxMaskCheckInput(const aMask   : string;
                            var Input   : string;
                                AutoFill: Boolean): integer;
begin
  if not (Picture(aMask, Input, AutoFill, result) in [prError, prSyntax]) then
    result := 0;
end;

function PxMaskValidate(const aMask : string;
                        const Input : string): integer;
var
  Str: String;
begin
  Str := Input;
  if Picture(aMask, Str, False, result) in [prComplete, prEmpty] then
    result := 0;
end;

function PxMaskValidateText(const aMask : string;
                            var Input : string): integer;
begin
  if Picture(aMask, Input, False, result) in [prComplete, prEmpty] then
    result := 0;
end;

{**** TPxMask ****}
type
  TSelection =
  record
    StartPos, EndPos: Integer;
  end;

constructor TPxMask.Create(aOwner : TComponent);
begin
  inherited Create;
  Clear;
end;

procedure TPxMask.SetAutoValidate(value: boolean);
begin
  if value <> AutoValidate then
  begin
    FAutoValidate := value;
    include(FState, pxStored);
  end;
end;

procedure TPxMask.SetAutoFill(value: boolean);
begin
  if value <> AutoFill then
  begin
    FAutoFill := value;
    include(FState, pxStored);
  end;
end;

procedure TPxMask.SetErrMsg(const value: string);
begin
  if value <> ErrorMessage then
  begin
    FErrMsg := value;
    include(FState, pxStored);
  end;
end;

procedure TPxMask.SetMask(const value : string);
var
  s : string;
begin
  if value <> Mask then
  begin
    if value <> '' then
    begin
      s := value;
      CheckSyntax(s);
    end;
    if (pxMaskFromParadox in FState) then
      if (csDesigning in Application{Owner}.ComponentState) then
        raise exception.create(SPxMaskDefinedOnTableLevel)
      else
        exit;
    include(FState, pxStored);
    FMask := s;
    if value <> '' then include(FState, pxModified);
  end;
end;

procedure TPxMask.SetRaiseInputError(value: boolean);
begin
  if value <> RaiseInputError then
  begin
    FRaiseInputError := value;
    include(FState, pxStored);
  end;
end;

function TPxMask.MaskStored : boolean;
begin
  result := FState * [pxMaskFromParadox, pxStored] = [pxStored];
end;

function TPxMask.PropsStored: boolean;
begin
  result := pxStored in FState;
end;

function TControlPxMask.GetTextLen: integer;
begin
  result := SendMessage(EditHandle, WM_GETTEXTLENGTH, 0, 0);
end;

function TControlPxMask.GetModified : boolean;
begin
  result := pxModified in FState;
end;

procedure TPxMask.Validate(var value : string);
begin
  if IsMasked then
  begin
    value := CheckValid(value);
    if ErrorPos = 0 then exit;
    ValidateError(ErrorPos > length(value));
    Abort;
  end;
end;

function TPxMask.CheckValid(s : string): string;
begin
  FErrPos := PxMaskValidateText(Mask, s);
  if (ErrorPos <> 0) then
  begin
    PxMaskCheckInput(Mask, s, AutoFill);
    FErrPos := PxMaskValidateText(Mask, s);
  end;
  if (ErrorPos <> 0) and not AutoFill then
  begin
    PxMaskCheckInput(Mask, s, true);
    FErrPos := PxMaskValidateText(Mask, s);
  end;
  result := s;
  ErrorPos := FErrPos;
end;

procedure TControlPxMask.CheckEditText;
begin
  CheckValid(EditText);
  EditModified := true;
end;

procedure TPxMask.Assign(Source: TPersistent);
var
  l: LongInt;
begin
  if not (Source is TPxMask) then
    if HasPxMask(Source, TPxMask(l)) and (TObject(l) is TPxMask) then
      Source := TPersistent(l);
  if Source is TPxMask then
  with TPxMask(Source) do begin
    if not (pxMaskFromParadox in self.FState) then
      self.Mask := Mask;
    self.ErrorMessage := ErrorMessage;
    self.AutoFill := AutoFill;
    self.AutoValidate := AutoValidate;
    self.RaiseInputError := RaiseInputError;
  end
  else
    if source <> nil then inherited Assign(Source);
end;

procedure TPxMask.LoadMask;
begin
  if MaskLoaded then exclude(FState, pxStored) else include(FState, pxStored);
end;

function TPxMask.MaskLoaded : boolean;
begin
  result := false;
end;

function TPxMask.FormatErrorText(Ident: string): string;
var
  i : integer;
const
  cMask = '%MASK%';
  cError = '%ERROR%';
begin
  result := ErrorMessage;
  if result <> '' then
  begin
    i := Pos(cError, result);
    if i > 0 then
      result := Format('%s%s%s', [Copy(result, 1, i - 1), Ident,
                                  Copy(result, i + Length(cError), MaxInt)]);
  end
  else
    result := Ident;
  i := Pos(cMask, result);
  if i > 0 then
    result := Format('%s%s%s', [Copy(result, 1, i - 1),
                                Mask,
                                Copy(result, i + Length(cMask), MaxInt)]);
end;

procedure TPxMask.Clear;
begin
  ErrorMessage := '';
  Mask := '';
  FAutoFill := AutoFillDefault;
  FAutoValidate := AutoValidateDefault;
  FRaiseInputError := RaiseInputErrorDefault;
end;

function TPxMask.Cleared: boolean;
begin
  result := (AutoFill = AutoFillDefault) and
            (AutoValidate = AutoValidateDefault) and
            (RaiseInputError = RaiseInputErrorDefault) and
            (Mask = '') and
            (ErrorMessage = '');
end;

procedure TPxMask.CheckSyntax(var aMask : string);
begin
  PxMaskCheckSyntax(aMask, FErrPos);
end;

function TPxMask.isMasked : boolean;
begin
  result := Mask <> '';
end;

procedure TPxMask.InputError;
var
  s : string;
  Error : EPxMaskError;
begin
  if BeepOnPxError and BeepOnPxInputError then Beep;
  if RaiseInputError then
  begin
    s := FormatErrorText(SPxMaskIllegalChar);
    if s <> '' then
    begin
      Error := EPxMaskError.create(s);
      Error.ErrorCode := errPxMaskIllegalChar;
      include(FState, pxReEnter);
      raise Error;
    end;
  end;
end;

procedure TPxMask.ValidChanged;
begin
end;

procedure TPxMask.ValidateError(Incomplete : boolean);
var
  s : string;
  Error : EPxMaskError;
begin
  if BeepOnPxError then Beep;
  if Incomplete then
    s := FormatErrorText(SPxMaskInputIncomplete)
  else
    s := FormatErrorText(SPxMaskIllegalString);
  if s <> '' then
  begin
    Error := EPxMaskError.create(s);
    if Incomplete then
      Error.ErrorCode := errPxMaskInputIncomplete
    else
      Error.ErrorCode := errPxMaskIllegalString;
    raise Error;
  end
  else
    abort;
end;

{*** TControlPxMask ***}
constructor TControlPxMask.create(aOwner : TComponent; EnableReset: boolean);
begin
  inherited create(aOwner);
  FOwner := aOwner;
  FEnableReset := EnableReset;
  FControlWndProc := (aOwner as TControl).WindowProc;
  TControl(aOwner).WindowProc := HookWndProc;
end;

destructor TControlPxMask.Destroy;
begin
  if Assigned(FControlWndProc) then TControl(Owner).WindowProc := FControlWndProc;
  inherited destroy;
end;

procedure TControlPxMask.SetModified(Value: Boolean);
begin
  if value then
    include(FState, pxModified)
  else
    exclude(FState, pxModified);
end;

function TControlPxMask.GetText: string;
var
  Len: Integer;
begin
  Len := GetTextLen;
  SetString(Result, PChar(nil), Len);
  if Len <> 0 then
    SendMessage(EditHandle, WM_GETTEXT, Len + 1, Longint(Result));
end;

procedure TControlPxMask.SetText(const value: string);
var
  EH : THandle;
begin
  if GetText <> Value then
  begin
    EH := EditHandle;
    SendMessage(EH, WM_SETTEXT, 0, Longint(PChar(Value)));
    SendMessage(EH, CM_TEXTCHANGED, 0, 0);
    CheckValid(value);
  end;
end;

procedure TControlPxMask.PrepareReset(const OldText : string);
begin
  if FEnableReset and not (csDesigning in FOwner.ComponentState) then
  begin
    if not (pxReEnter in FState) then
      FOldValue := OldText;
  end;
  EditModified := false;
  CheckValid(EditText);
end;

procedure TControlPxMask.HookWndProc(var Message: TMessage);

  function DoKeyPress: boolean;
  var
    Key: Char;
  begin
    with TWMChar(Message) do begin
      Key := Char(CharCode);
      CheckResetKey(Key);
      HookKeyPress(Key);
      CharCode := Word(Key);
    end;
    result := (Key = #0);
  end;

begin
  with Message do begin
    case Msg of
      WM_CHAR: if DoKeyPress then exit;
      CM_Exit:
        if AutoValidate and IsMasked and not (csDesigning in FOwner.ComponentState) then
          ValidateEdit;
      CM_Enter:
        if IsMasked and not (csDesigning in FOwner.ComponentState) then
        begin
          if not (pxReEnter in FState) then
            PrepareReset(EditText)
          else
          begin
            SetCaret;
            exclude(FState, pxReEnter);
            exit;
          end
        end;
    end;

    Dispatch(Message);

    case Msg of
      WM_Command:
        with TWMCommand(Message) do
          if Ctl = EditHandle then
            case NotifyCode of
              EN_SETFOCUS: if (ErrorPos > 0) then SetCaret;
              EN_CHANGE:   CheckEditText;
            end;
      CN_Command:
        with TWMCommand(message) do
        case NotifyCode of
          CBN_SELCHANGE : CheckEditText;
        end;
      WM_KeyUp:
        with TWMKeyUp(message) do
          if ((CharCode = VK_DELETE) and
              ([ssShift, ssCtrl] * KeyDataToShiftState(KeyData) = [])) or
             (CharCode = VK_BACK)
          then
            CheckEditText
          else
            if (CharCode = VK_ESCAPE) then
              ErrorPos := 0;
      WM_Cut,
      WM_Paste:
        if isMasked then CheckEditText;
      CM_TextChanged:
        if not (Owner is TCustomComboBox) then
          PrepareReset(EditText);
      CM_WantSpecialKey:
        if (TWMKey(Message).CharCode = VK_ESCAPE) and IsMasked and EditModified then
          Message.Result := 1;
    end;
  end;
end;

procedure TControlPxMask.SetCaret;
begin
  if ErrorPos = 0 then exit;
  SendMessage(EditHandle, EM_SETSEL, ErrorPos - 1, ErrorPos - 1);
end;

procedure TControlPxMask.CheckResetKey(var Key : char);
begin
  if not (pxSending in FState) and (cardinal(key) = VK_ESCAPE) then
  begin
    if EditModified then
    begin
      Reset;
      key := #0;
    end;
    ErrorPos := 0;
  end;
end;

procedure TPxMask.SetErrorPos(value : integer);
var
  prevValid: boolean;
begin
  FErrPos := value;
  prevValid := (pxValid in FState);
  if value > 0 then exclude(FState, pxValid) else include(FState, pxValid);
  if prevValid xor (pxValid in FState) then
    ValidChanged;
end;

function TControlPxMask.GetEditControl: TWinControl;
begin
  result := FindControl(EditHandle);
  if result = nil then
    result := TWinControl(Owner);
end;

function TControlPxMask.GetMaxLen: integer;
var
  Ctl: TWinControl;
begin
  result := 0;
  Ctl := EditControl;
  if (Ctl is TCustomEdit) then
    result := TPxMaskEdit(Ctl).MaxLength
  else
    if (Ctl is TCustomComboBox) then
      result := TPxComboBox(Ctl).MaxLength;
end;

procedure TControlPxMask.HookKeyPress(var Key: Char);
var
  s: string;
  MaxLen, i, len, errpos: Integer;
  Selection: TSelection;
  EH : THandle;
begin
  if (Key in [#32..#255]) and (not (pxSending in FState)) and
     IsMasked and CanEditModify
  then
  begin
    EH := EditHandle;
    SendMessage(EH, EM_GETSEL, Longint(@Selection.StartPos), Longint(@Selection.EndPos));
    SetLength(s, Selection.StartPos + 1);
    SendMessage(EH, WM_GETTEXT, succ(Selection.StartPos), longint(s));
    s[succ(Selection.StartPos)] := Key;
    Key := #0;
    if Selection.EndPos = SendMessage(EH, WM_GETTEXTLENGTH, 0, 0) then
    begin
      ErrPos := PxMaskCheckInput(Mask, s, AutoFill);
      len := length(s) - Selection.StartPos;
    end
    else
    begin
      len := 1;
      ErrPos := PxMaskCheckInput(Mask, s, false);
    end;
    if ErrPos = 0 then
    begin
      MaxLen := 0;
      if (len > 1) then
      begin
        MaxLen := EditMaxLen;
        if (MaxLen > 0) and (length(s) > MaxLen) then
          SendMessage(EH, EM_LIMITTEXT, length(s), 0)
        else
          MaxLen := 0;
      end;
      include(FState, pxSending);
      try
        for i := Selection.StartPos + 1 to Selection.StartPos + len do
          SendMessage(EH, WM_CHAR, ord(s[i]), 1);
      finally
        exclude(FState, pxSending);
        if MaxLen > 0 then
          SendMessage(EH, EM_LIMITTEXT, MaxLen, 0)
      end;
      CheckEditText;
    end
    else
    begin
      CheckEditText;
      InputError;
      abort;
    end;
  end
  else
    if (cardinal(key) = VK_RETURN) and AutoValidate then
      validateEdit;
end;

procedure TControlPxMask.ValidateEdit;
var
  s: string;
begin
  if EditModified then
  begin
    s := EditText;
    Validate(s);
    EditText := s;
  end;
end;

procedure TControlPxMask.ValidateError(Incomplete : boolean);
begin
  if not (csDesigning in FOwner.ComponentState) and (Owner is TWinControl) then
  begin
    include(FState, pxReEnter);
    with TWinControl(Owner) do
      if CanFocus then SetFocus;
  end;
  inherited;
end;

procedure TControlPxMask.Reset;
begin
  if FEnableReset and not (csDesigning in FOwner.ComponentState) and
    EditModified then
  begin
    EditText := FOldValue;
    SendMessage(EditHandle, EM_SETSEL, 0, -1);
  end;
  exclude(FState, pxReEnter);
  EditModified := false;
  CheckValid(EditText);
end;

function TControlPxMask.CanEditModify: boolean;
begin
  result := false;
  if (Owner is TCustomEdit) then
    result := not TPxMaskEdit(Owner).ReadOnly
  else
    if (Owner is TCustomComboBox) then
      result := TPxComboBox(Owner).Style < csDropDownList;
end;

procedure TControlPxMask.ValidChanged;
begin
  if assigned(FOnValidChanged) then
    FOnValidChanged(self, EditControl, ErrorPos = 0);
end;

procedure TControlPxMask.DefaultHandler(var Message);
begin
  FControlWndProc(TMessage(Message));
end;

function TControlPxMask.GetEditHandle: THandle;
begin
  result := 0;
  if TWinControl(FOwner).HandleAllocated then
  begin
    result := TWinControl(FOwner).Handle;
    if FOwner is TCustomComboBox then
      result := TPxComboBox(FOwner).EditHandle;
  end;
end;

{*** TPxMaskEdit ***}
constructor TPxMaskEdit.create(aOwner : TComponent);
begin
  inherited create(aOwner);
  FPxMask := TControlPxMask.create(self, true); {//MUST be created after inherited}
end;

destructor TPxMaskEdit.Destroy;
begin
  if assigned(FPxMask) then FPxMask.free;
  inherited Destroy;
end;

function TPxMaskEdit.GetOnValidChanged: TValidChangedEvent;
begin
  result := FPxMask.OnValidChanged;
end;

procedure TPxMaskEdit.SetOnValidChanged(value: TValidChangedEvent);
begin
  FPxMask.OnValidChanged := value;
end;

procedure TPxMaskEdit.SetPxMask(value : TControlPxMask);
begin
  FPxMask.assign(value);
end;

{*** TPxComboBox ***}
constructor TPxComboBox.create(aOwner : TComponent);
begin
  inherited create(aOwner);
  FPxMask := TControlPxMask.create(self, true); {//MUST be created after inherited}
end;

destructor TPxComboBox.Destroy;
begin
  if assigned(FPxMask) then FPxMask.free;
  inherited Destroy;
end;

function TPxComboBox.GetOnValidChanged: TValidChangedEvent;
begin
  result := FPxMask.OnValidChanged;
end;

procedure TPxComboBox.SetOnValidChanged(value: TValidChangedEvent);
begin
  FPxMask.OnValidChanged := value;
end;

procedure TPxComboBox.KeyPress(var Key: Char);
begin
  {ComboBoxes don't get WM_CHAR messages (they're redirected to the EditHandle),
   so you've to handle them in KeyPress}
  FPxMask.CheckResetKey(Key);
  FPxMask.HookKeyPress(Key);
  inherited KeyPress(Key);
end;

procedure TPxComboBox.SetPxMask(value : TControlPxMask);
begin
  FPxMask.assign(value);
end;

function HasPxMask(Control: TPersistent; var PxMask: TPxMask) : boolean;
begin
  result := GetOrdProperty(Control, 'PxMask', tkClass, LongInt(PxMask));
end;

end.

